from typing import List

from pyelectric.electronic.amplifier.input import Input as _Input


class Input(_Input):
    Vcc: float
    Rc: float
    beta: float
    ro: float
    Rb: float

    def __init__(self, Vcc: float, Rc: float, beta: float, ro: float, Rb: float) -> None:
        self.Vcc = Vcc
        self.Rc = Rc
        self.beta = beta
        self.ro = ro
        self.Rb = Rb

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['Vcc', 'beta', 'ro', 'Rb', 'Rc']



from schemdraw import Drawing
from schemdraw import elements as elm

from .input import Input


def draw_void():
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.BjtNpn().right())

    drawing.push()
    drawing += elm.Resistor().up().label('Rc')
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().reverse().label('Vcc')
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label('R1')
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.base)

    drawing.pop()
    drawing.push()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor().left().reverse().label('C1')
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing.pop()
    drawing += elm.Resistor().label('R2')
    drawing += elm.Ground()

    drawing += elm.Resistor().label('Re').down().at(transistor.emitter).length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.emitter)
    drawing += elm.Line().right().length(0.75*drawing.unit)
    drawing += elm.Capacitor().label('Ce').down().length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Dot()
    drawing += elm.Capacitor().right().at(transistor.collector).label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing


def draw(amplifier_input: Input):
    R2 = str(round(amplifier_input.R2, 2))
    R1 = str(round(amplifier_input.R1, 2))
    Rc = str(round(amplifier_input.Rc, 2))
    Re = str(round(amplifier_input.Re, 2))
    Vcc = str(round(amplifier_input.Vcc, 2))

    drawing = Drawing()

    transistor = drawing.add(elm.transistors.BjtNpn().right())

    drawing.push()
    drawing += elm.Resistor().up().label(Rc)
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().reverse().label(Vcc)
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label(R1)
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.base)

    drawing.pop()
    drawing.push()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor().left().reverse().label('C1')
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing.pop()
    drawing += elm.Resistor().label(R2)
    drawing += elm.Ground()

    drawing += elm.Resistor().label(Re).down().at(transistor.emitter).length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.emitter)
    drawing += elm.Line().right().length(0.75*drawing.unit)
    drawing += elm.Capacitor().label('Ce').down().length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Dot()
    drawing += elm.Capacitor().right().at(transistor.collector).label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing



from typing import List

from pyelectric.electronic.amplifier.input import Input as _Input


class Input(_Input):
    Vdd: float
    Vgg: float
    Vp: float
    Idss: float
    rd: float
    Rd: float
    Rg: float

    def __init__(self, Vdd: float, Vgg: float, Vp: float, Idss: float, rd: float, Rd: float, Rg: float):
        self.Vdd = Vdd
        self.Vgg = Vgg
        self.Vp = Vp
        self.Idss = Idss
        self.rd = rd
        self.Rd = Rd
        self.Rg = Rg

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['Vdd', 'Vgg', 'Vp', 'Idss', 'rd', 'Rd', 'Rg']



import sympy as sp


class Step:
    equation: sp.Eq

    def __init__(self, a, b) -> None:
        if isinstance(a, str):
            a = sp.Symbol(a)
        if isinstance(b, str):
            b = sp.Symbol(b)
        self.equation = sp.Eq(a, b, evaluate=False)

    def __str__(self) -> str:
        return str(self.equation)



import math
from typing import Tuple


def cutoff_frequencies(R: float, L: float, C: float) -> Tuple[float, float]:
    f1 = (1/(2*math.pi))*(-(R/(2*L)) + (1/2)*math.sqrt((R/L)**2 + (4/L*C)))
    f2 = (1/(2*math.pi))*((R/(2*L)) + (1/2)*math.sqrt((R/L)**2 + (4/L*C)))
    return f1, f2



import math
from functools import reduce
from typing import List


class Position:
    x: float
    y: float

    def __init__(self, x: float, y: float, polar: bool = False):
        if polar:
            self.x = x * math.cos(y)
            self.y = x * math.sin(y)
        else:
            self.x = x
            self.y = y

    def __str__(self) -> str:
        return f"({self.x}, {self.y})"

    def __repr__(self) -> str:
        return f"Position({self.x}, {self.y})"

    def __getitem__(self, item: int) -> float:
        if item == 0:
            return self.x
        elif item == 1:
            return self.y
        else:
            raise IndexError("Index out of range")


def geometric_mean_position(positions: List[Position]) -> Position:
    x_list = [position.x for position in positions]
    y_list = [position.y for position in positions]
    x = reduce(lambda a, b: a*b, x_list)**(1/len(x_list))
    y = reduce(lambda a, b: a*b, y_list)**(1/len(y_list))
    return Position(x, y)



from typing import List

from pyelectric.electronic.amplifier.input import Input as _Input


class Input(_Input):
    Vdd: float
    Rg: float
    Rs: float
    Vp: float
    Idss: float
    rd: float
    Idq: float

    def __init__(self, Vdd: float, Rg: float, Rs: float, Vp: float, Idss: float, rd: float, Idq: float):
        self.Vdd = Vdd
        self.Rg = Rg
        self.Rs = Rs
        self.Vp = Vp
        self.Idss = Idss
        self.rd = rd
        self.Idq = Idq

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['Vdd', 'Vp', 'Idss', 'rd', 'Rg', 'Rs', 'Idq']



from termcolor import colored


def complex_to_str(c: complex) -> str:
    return f'{c.real:.4e} + {c.imag:.4e}j'


class Bar:
    name: str
    voltage: complex = 0 + 0j
    power: complex = 0 + 0j

    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        V = colored(complex_to_str(self.voltage), 'green')
        S = colored(complex_to_str(self.power), 'green')
        bar = colored(self.name, 'blue')
        return f'{bar}: V = ({V}), S = ({S})'



from typing import List

from src.models.amplifier.output import Output as _Output


class Output(_Output):
    re: float
    Zi: float
    Zo: float
    Avnl: float

    def __init__(self, re: float, Zi: float, Zo: float, Avnl: float) -> None:
        self.re = re
        self.Zi = Zi
        self.Zo = Zo
        self.Avnl = Avnl

    def __str__(self) -> str:
        return '\n'.join([
            f"re = {self.re:.2f} Ω",
            f"Zi = {self.Zi:.2f} Ω",
            f"Zo = {self.Zo:.2f} Ω",
            f"Avnl = {self.Avnl:.2f}"
        ])

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['re', 'Zi', 'Zo', 'Avnl']



import math
from copy import deepcopy
from functools import reduce
from typing import Any, List

from .conductor import Conductor


def list_of_lists_to_list(list_of_list: List[List[Any]]) -> List[Any]:
    return [item for sublist in list_of_list for item in sublist]


def geometric_mean(list_of_numbers: List[float]) -> float:
    return reduce(lambda x, y: x*y, list_of_numbers)**(1/len(list_of_numbers))


def geometric_mean_distance(conductors_bundles: List[List[Conductor]]) -> float:
    distance: float = 1
    n: int = 0
    for i, conductors_bundle in enumerate(conductors_bundles):
        compare_conductors_bundles = conductors_bundles[i+1:]
        compare_conductors = list_of_lists_to_list(compare_conductors_bundles)
        for conductor1 in conductors_bundle:
            for conductor2 in compare_conductors:
                d = conductor1.distance(conductor2.position)
                distance *= d if d > 0 else conductor1.radius
                n += 1
    distance = distance**(1/n)
    return distance


def get_discounted_arrow_conductors(conductors_bundles: List[List[Conductor]], arrow: float) -> List[List[Conductor]]:
    conductors_bundles = deepcopy(conductors_bundles)
    for conductors_bundle in conductors_bundles:
        for conductor in conductors_bundle:
            conductor.position.y -= (2/3)*arrow
    return conductors_bundles


def get_ground_conductors(conductors_bundles: List[List[Conductor]]) -> List[List[Conductor]]:
    conductors_bundles = deepcopy(conductors_bundles)
    for conductors_bundle in conductors_bundles:
        for conductor in conductors_bundle:
            conductor.position.y *= -1
    return conductors_bundles


def calc_alpha(conductors_bundles: List[List[Conductor]], arrow: float) -> float:
    up_conductors_bundles = get_discounted_arrow_conductors(
        conductors_bundles, arrow)
    down_conductors_bundles = get_ground_conductors(up_conductors_bundles)

    Hs_list = [geometric_mean_distance([up, down]) for up, down in zip(
        up_conductors_bundles, down_conductors_bundles)]
    Hs = geometric_mean(Hs_list)

    Hm = geometric_mean_distance([
        list_of_lists_to_list(up_conductors_bundles),
        list_of_lists_to_list(down_conductors_bundles)
    ])

    alpha = math.log(Hm/Hs)

    return alpha


def calc_capacitance(conductors_bundles: List[List[Conductor]], arrow: float = None) -> float:
    Dm = geometric_mean_distance(conductors_bundles)
    Ds_list = [geometric_mean_distance([c, c]) for c in conductors_bundles]

    alpha: float = 0 if arrow is None else calc_alpha(
        conductors_bundles, arrow)
    e0 = 8.85e-12
    C_list = [2*math.pi*e0/(math.log(Dm/Ds) - alpha) for Ds in Ds_list]
    C = sum(C_list)/len(conductors_bundles)
    return C



import sympy as sp
from pyelectric.electronic.amplifier import Amplifier as _Amplifier

from . import drawing
from .input import Input
from .output import Output


class Amplifier(_Amplifier):
    input: Input = Input(0, 0, 0, 0, 0, 0, 0, 0)
    output: Output = Output(0, 0, 0, 0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.input = Input(*args, **kwargs)

    def __call__(self) -> Output:
        Vdd = self.input.Vdd
        R1 = self.input.R1
        R2 = self.input.R2
        Rd = self.input.Rd
        Rs = self.input.Rs
        Vp = self.input.Vp
        Idss = self.input.Idss
        rd = self.input.rd

        Vg = Vdd*R2/(R1+R2)

        Vgs = sp.Symbol('Vgs')
        equation = Vg - Rs*Idss*(1 - Vgs/Vp)**2 - Vgs
        Vgs1, Vgs2 = sp.solve(equation, Vgs)
        Vgs1, Vgs2 = float(Vgs1), float(Vgs2)
        Vgs1, Vgs2
        Vgs = Vgs1 if Vgs1 < 0 and Vgs1 > Vp else Vgs2

        gm = (2*Idss/abs(Vp))*(1 - (Vgs/Vp))

        Zi = 1/(1/R1 + 1/R2)
        Zo = 1/((1/rd) + (1/Rd))
        Avnl = -gm*Zo
        return Output(gm, Zi, Zo, Avnl)

    @staticmethod
    def draw_void():
        return drawing.draw_void()

    def draw(self):
        return drawing.draw(self.input)
